# coding: utf-8

from __future__ import print_function, absolute_import, division, unicode_literals

if False:  # MYPY
    from typing import Dict, Any  # NOQA

_package_data = dict(
    full_package_name="strictyaml.ruamel",
    version_info=(0, 16, 13),
    __version__="0.16.13",
    author="Anthon van der Neut",
    author_email="a.van.der.neut@ruamel.eu",
    description="strictyaml.ruamel is a YAML parser/emitter that supports roundtrip preservation of comments, seq/map flow style, and map key order",  # NOQA
    entry_points=None,
    since=2014,
    extras_require={
        ':platform_python_implementation=="CPython" and python_version<="2.7"': [
            "ruamel.ordereddict"
        ],  # NOQA
        ':platform_python_implementation=="CPython" and python_version<"3.10"': [
            "strictyaml.ruamel.clib>=0.1.2"
        ],  # NOQA
        "jinja2": ["strictyaml.ruamel.jinja2>=0.2"],
        "docs": ["ryd"],
    },
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: Jython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup",
        "Typing :: Typed",
    ],
    keywords="yaml 1.2 parser round-trip preserve quotes order config",
    read_the_docs="yaml",
    supported=[(2, 7), (3, 5)],  # minimum
    tox=dict(
        env="*",  # remove 'pn', no longer test narrow Python 2.7 for unicode patterns and PyPy
        deps="ruamel.std.pathlib",
        fl8excl="_test/lib",
    ),
    universal=True,
    rtfd="yaml",
)  # type: Dict[Any, Any]


version_info = _package_data["version_info"]
__version__ = _package_data["__version__"]

try:
    from .cyaml import *  # NOQA

    __with_libyaml__ = True
except (ImportError, ValueError):  # for Jython
    __with_libyaml__ = False

from strictyaml.ruamel.main import *  # NOQA



# coding: utf-8

# Abstract classes.

if False:  # MYPY
    from typing import Any, Dict, Optional, List  # NOQA


def CommentCheck():
    # type: () -> None
    pass


class Event(object):
    __slots__ = "start_mark", "end_mark", "comment"

    def __init__(self, start_mark=None, end_mark=None, comment=CommentCheck):
        # type: (Any, Any, Any) -> None
        self.start_mark = start_mark
        self.end_mark = end_mark
        # assert comment is not CommentCheck
        if comment is CommentCheck:
            comment = None
        self.comment = comment

    def __repr__(self):
        # type: () -> Any
        attributes = [
            key
            for key in ["anchor", "tag", "implicit", "value", "flow_style", "style"]
            if hasattr(self, key)
        ]
        arguments = ", ".join(
            ["%s=%r" % (key, getattr(self, key)) for key in attributes]
        )
        if self.comment not in [None, CommentCheck]:
            arguments += ", comment={!r}".format(self.comment)
        return "%s(%s)" % (self.__class__.__name__, arguments)


class NodeEvent(Event):
    __slots__ = ("anchor",)

    def __init__(self, anchor, start_mark=None, end_mark=None, comment=None):
        # type: (Any, Any, Any, Any) -> None
        Event.__init__(self, start_mark, end_mark, comment)
        self.anchor = anchor


class CollectionStartEvent(NodeEvent):
    __slots__ = "tag", "implicit", "flow_style", "nr_items"

    def __init__(
        self,
        anchor,
        tag,
        implicit,
        start_mark=None,
        end_mark=None,
        flow_style=None,
        comment=None,
        nr_items=None,
    ):
        # type: (Any, Any, Any, Any, Any, Any, Any, Optional[int]) -> None
        NodeEvent.__init__(self, anchor, start_mark, end_mark, comment)
        self.tag = tag
        self.implicit = implicit
        self.flow_style = flow_style
        self.nr_items = nr_items


class CollectionEndEvent(Event):
    __slots__ = ()


# Implementations.


class StreamStartEvent(Event):
    __slots__ = ("encoding",)

    def __init__(self, start_mark=None, end_mark=None, encoding=None, comment=None):
        # type: (Any, Any, Any, Any) -> None
        Event.__init__(self, start_mark, end_mark, comment)
        self.encoding = encoding


class StreamEndEvent(Event):
    __slots__ = ()


class DocumentStartEvent(Event):
    __slots__ = "explicit", "version", "tags"

    def __init__(
        self,
        start_mark=None,
        end_mark=None,
        explicit=None,
        version=None,
        tags=None,
        comment=None,
    ):
        # type: (Any, Any, Any, Any, Any, Any) -> None
        Event.__init__(self, start_mark, end_mark, comment)
        self.explicit = explicit
        self.version = version
        self.tags = tags


class DocumentEndEvent(Event):
    __slots__ = ("explicit",)

    def __init__(self, start_mark=None, end_mark=None, explicit=None, comment=None):
        # type: (Any, Any, Any, Any) -> None
        Event.__init__(self, start_mark, end_mark, comment)
        self.explicit = explicit


class AliasEvent(NodeEvent):
    __slots__ = ()


class ScalarEvent(NodeEvent):
    __slots__ = "tag", "implicit", "value", "style"

    def __init__(
        self,
        anchor,
        tag,
        implicit,
        value,
        start_mark=None,
        end_mark=None,
        style=None,
        comment=None,
    ):
        # type: (Any, Any, Any, Any, Any, Any, Any, Any) -> None
        NodeEvent.__init__(self, anchor, start_mark, end_mark, comment)
        self.tag = tag
        self.implicit = implicit
        self.value = value
        self.style = style


class SequenceStartEvent(CollectionStartEvent):
    __slots__ = ()


class SequenceEndEvent(CollectionEndEvent):
    __slots__ = ()


class MappingStartEvent(CollectionStartEvent):
    __slots__ = ()


class MappingEndEvent(CollectionEndEvent):
    __slots__ = ()



from strictyaml.exceptions import YAMLSerializationError, InvalidOptionalDefault
from strictyaml.validators import Validator, MapValidator, SeqValidator
from strictyaml.ruamel.comments import CommentedMap, CommentedSeq
from strictyaml.representation import YAML
from strictyaml.scalar import ScalarValidator, Str
from strictyaml.yamllocation import YAMLChunk
import sys


if sys.version_info[0] == 3:
    unicode = str


class Optional(object):
    def __init__(self, key, default=None, drop_if_none=True):
        self.key = key
        self.default = default
        self.drop_if_none = drop_if_none

    def __repr__(self):
        # TODO: Add default
        return 'Optional("{0}")'.format(self.key)


class MapPattern(MapValidator):
    def __init__(
        self, key_validator, value_validator, minimum_keys=None, maximum_keys=None
    ):
        self._key_validator = key_validator
        self._value_validator = value_validator
        self._maximum_keys = maximum_keys
        self._minimum_keys = minimum_keys
        assert isinstance(
            self._key_validator, ScalarValidator
        ), "key_validator must be ScalarValidator"
        assert isinstance(
            self._value_validator, Validator
        ), "value_validator must be Validator"
        assert isinstance(
            maximum_keys, (type(None), int)
        ), "maximum_keys must be an integer"
        assert isinstance(
            minimum_keys, (type(None), int)
        ), "maximum_keys must be an integer"

    @property
    def key_validator(self):
        return self._key_validator

    def validate(self, chunk):
        items = chunk.expect_mapping()

        if self._maximum_keys is not None and len(items) > self._maximum_keys:
            chunk.expecting_but_found(
                "while parsing a mapping",
                "expected a maximum of {0} key{1}, found {2}.".format(
                    self._maximum_keys,
                    "s" if self._maximum_keys > 1 else "",
                    len(items),
                ),
            )

        if self._minimum_keys is not None and len(items) < self._minimum_keys:
            chunk.expecting_but_found(
                "while parsing a mapping",
                "expected a minimum of {0} key{1}, found {2}.".format(
                    self._minimum_keys,
                    "s" if self._minimum_keys > 1 else "",
                    len(items),
                ),
            )

        for key, value in items:
            yaml_key = self._key_validator(key)
            key.process(yaml_key)
            value.process(self._value_validator(value))
            chunk.add_key_association(key.contents, yaml_key.data)

    def to_yaml(self, data):
        self._should_be_mapping(data)
        # TODO : Maximum minimum keys
        return CommentedMap(
            [
                (self._key_validator.to_yaml(key), self._value_validator.to_yaml(value))
                for key, value in data.items()
            ]
        )

    def __repr__(self):
        return "MapPattern({0}, {1})".format(
            repr(self._key_validator), repr(self._value_validator)
        )


class Map(MapValidator):
    def __init__(self, validator, key_validator=None):
        self._validator = validator
        self._key_validator = Str() if key_validator is None else key_validator
        assert isinstance(
            self._key_validator, ScalarValidator
        ), "key validator must be ScalarValidator"

        self._validator_dict = {
            key.key if isinstance(key, Optional) else key: value
            for key, value in validator.items()
        }

        self._required_keys = [
            key for key in validator.keys() if not isinstance(key, Optional)
        ]

        for key_val, value_val in validator.items():
            if isinstance(key_val, Optional):
                if key_val.default is not None and not key_val.drop_if_none:
                    raise InvalidOptionalDefault(
                        "If you have a default that isn't None, drop_if_none must be True."
                    )
                if key_val.default is not None and key_val.drop_if_none:
                    try:
                        value_val.to_yaml(key_val.default)
                    except YAMLSerializationError as error:
                        raise InvalidOptionalDefault(
                            "Optional default for '{}' failed validation:\n  {}".format(
                                key_val.key, error
                            )
                        )

        self._defaults = {
            key.key: key.default
            for key in validator.keys()
            if isinstance(key, Optional)
            and (key.default is not None or not key.drop_if_none)
        }

    @property
    def key_validator(self):
        return self._key_validator

    def __repr__(self):
        # TODO : repr key_validator
        return "Map({{{0}}})".format(
            ", ".join(
                [
                    "{0}: {1}".format(repr(key), repr(value))
                    for key, value in self._validator.items()
                ]
            )
        )

    def get_validator(self, key):
        return self._validator_dict[key]

    def unexpected_key(self, key, yaml_key, value, chunk):
        key.expecting_but_found(
            "while parsing a mapping",
            "unexpected key not in schema '{0}'".format(unicode(yaml_key.scalar)),
        )

    def validate(self, chunk):
        found_keys = set()
        items = chunk.expect_mapping()

        for key, value in items:
            yaml_key = self._key_validator(key)

            if yaml_key.scalar not in self._validator_dict.keys():
                self.unexpected_key(key, yaml_key, value, chunk)

            value.process(self.get_validator(yaml_key.scalar)(value))
            key.process(yaml_key)
            chunk.add_key_association(key.contents, yaml_key.data)
            found_keys.add(yaml_key.scalar)

        for default_key, default_data in self._defaults.items():
            if default_key not in [key.contents for key, _ in items]:
                key_chunk = YAMLChunk(default_key)
                yaml_key = self._key_validator(key_chunk)
                strictindex = yaml_key.data
                value_validator = self.get_validator(default_key)
                new_value = value_validator(
                    YAMLChunk(value_validator.to_yaml(default_data))
                )
                forked_chunk = chunk.fork(strictindex, new_value)
                forked_chunk.val(strictindex).process(new_value)
                updated_value = value_validator(forked_chunk.val(strictindex))
                updated_value._chunk.make_child_of(chunk.val(strictindex))
                # marked_up = new_value.as_marked_up()
                # chunk.contents[chunk.ruamelindex(strictindex)] = marked_up
                chunk.add_key_association(default_key, strictindex)
                sp = chunk.strictparsed()
                if isinstance(sp, YAML):
                    # Do not trigger __setitem__ validation at this point, as
                    # we just ran the validator, and
                    # representation.py:revalidate() doesn't overwrite the
                    # _validator property until after all values are checked,
                    # which leads to an exception being raised if it is
                    # re-checked.
                    sp._value[yaml_key] = updated_value
                else:
                    sp[yaml_key] = updated_value

        if not set(self._required_keys).issubset(found_keys):
            chunk.while_parsing_found(
                "a mapping",
                "required key(s) '{0}' not found".format(
                    "', '".join(
                        sorted(list(set(self._required_keys).difference(found_keys)))
                    )
                ),
            )

    def to_yaml(self, data):
        self._should_be_mapping(data)
        # TODO : if keys not in list or required keys missing, raise exception.
        return CommentedMap(
            [
                (key, self.get_validator(key).to_yaml(value))
                for key, value in data.items()
                if key not in self._defaults.keys()
                or key in self._defaults.keys()
                and value != self._defaults[key]
            ]
        )


class MapCombined(Map):
    def __init__(self, map_validator, key_validator, value_validator):
        super(MapCombined, self).__init__(map_validator, key_validator)
        self._value_validator = value_validator

    def get_validator(self, key):
        return self._validator_dict.get(key, self._value_validator)

    def unexpected_key(self, key, yaml_key, value, chunk):
        pass


class Seq(SeqValidator):
    def __init__(self, validator):
        self._validator = validator

    def __repr__(self):
        return "Seq({0})".format(repr(self._validator))

    def validate(self, chunk):
        for item in chunk.expect_sequence():
            item.process(self._validator(item))

    def to_yaml(self, data):
        self._should_be_list(data)
        return CommentedSeq([self._validator.to_yaml(item) for item in data])


class FixedSeq(SeqValidator):
    def __init__(self, validators):
        self._validators = validators
        for item in validators:
            assert isinstance(
                item, Validator
            ), "all FixedSeq validators must be Validators"

    def __repr__(self):
        return "FixedSeq({0})".format(repr(self._validators))

    def validate(self, chunk):
        sequence = chunk.expect_sequence(
            "when expecting a sequence of {0} elements".format(len(self._validators))
        )

        if len(self._validators) != len(sequence):
            chunk.expecting_but_found(
                "when expecting a sequence of {0} elements".format(
                    len(self._validators)
                ),
                "found a sequence of {0} elements".format(len(chunk.contents)),
            )

        for item, validator in zip(sequence, self._validators):
            item.process(validator(item))

    def to_yaml(self, data):
        self._should_be_list(data)
        # TODO : Different length string
        return CommentedSeq(
            [validator.to_yaml(item) for item, validator in zip(data, self._validators)]
        )


class UniqueSeq(SeqValidator):
    def __init__(self, validator):
        self._validator = validator
        assert isinstance(
            self._validator, ScalarValidator
        ), "UniqueSeq validator must be ScalarValidator"

    def __repr__(self):
        return "UniqueSeq({0})".format(repr(self._validator))

    def validate(self, chunk):
        existing_items = set()

        for item in chunk.expect_sequence("when expecting a unique sequence"):
            if item.contents in existing_items:
                chunk.while_parsing_found("a sequence", "duplicate found")
            else:
                existing_items.add(item.contents)
                item.process(self._validator(item))

    def to_yaml(self, data):
        self._should_be_list(data)

        if len(set(data)) < len(data):
            raise YAMLSerializationError(
                (
                    "Expecting all unique items, "
                    "but duplicates were found in '{}'.".format(data)
                )
            )

        return CommentedSeq([self._validator.to_yaml(item) for item in data])



# # header
# coding: utf-8

from __future__ import unicode_literals

if False:  # MYPY
    from typing import Text, Any, Dict, Optional, List  # NOQA
    from .error import StreamMark  # NOQA

SHOWLINES = True


class Token(object):
    __slots__ = "start_mark", "end_mark", "_comment"

    def __init__(self, start_mark, end_mark):
        # type: (StreamMark, StreamMark) -> None
        self.start_mark = start_mark
        self.end_mark = end_mark

    def __repr__(self):
        # type: () -> Any
        # attributes = [key for key in self.__slots__ if not key.endswith('_mark') and
        #               hasattr('self', key)]
        attributes = [key for key in self.__slots__ if not key.endswith("_mark")]
        attributes.sort()
        arguments = ", ".join(
            ["%s=%r" % (key, getattr(self, key)) for key in attributes]
        )
        if SHOWLINES:
            try:
                arguments += ", line: " + str(self.start_mark.line)
            except:  # NOQA
                pass
        try:
            arguments += ", comment: " + str(self._comment)
        except:  # NOQA
            pass
        return "{}({})".format(self.__class__.__name__, arguments)

    def add_post_comment(self, comment):
        # type: (Any) -> None
        if not hasattr(self, "_comment"):
            self._comment = [None, None]
        self._comment[0] = comment

    def add_pre_comments(self, comments):
        # type: (Any) -> None
        if not hasattr(self, "_comment"):
            self._comment = [None, None]
        assert self._comment[1] is None
        self._comment[1] = comments

    def get_comment(self):
        # type: () -> Any
        return getattr(self, "_comment", None)

    @property
    def comment(self):
        # type: () -> Any
        return getattr(self, "_comment", None)

    def move_comment(self, target, empty=False):
        # type: (Any, bool) -> Any
        """move a comment from this token to target (normally next token)
        used to combine e.g. comments before a BlockEntryToken to the
        ScalarToken that follows it
        empty is a special for empty values -> comment after key
        """
        c = self.comment
        if c is None:
            return
        # don't push beyond last element
        if isinstance(target, (StreamEndToken, DocumentStartToken)):
            return
        delattr(self, "_comment")
        tc = target.comment
        if not tc:  # target comment, just insert
            # special for empty value in key: value issue 25
            if empty:
                c = [c[0], c[1], None, None, c[0]]
            target._comment = c
            # nprint('mco2:', self, target, target.comment, empty)
            return self
        if c[0] and tc[0] or c[1] and tc[1]:
            raise NotImplementedError("overlap in comment %r %r" % (c, tc))
        if c[0]:
            tc[0] = c[0]
        if c[1]:
            tc[1] = c[1]
        return self

    def split_comment(self):
        # type: () -> Any
        """split the post part of a comment, and return it
        as comment to be added. Delete second part if [None, None]
         abc:  # this goes to sequence
           # this goes to first element
           - first element
        """
        comment = self.comment
        if comment is None or comment[0] is None:
            return None  # nothing to do
        ret_val = [comment[0], None]
        if comment[1] is None:
            delattr(self, "_comment")
        return ret_val


# class BOMToken(Token):
#     id = '<byte order mark>'


class DirectiveToken(Token):
    __slots__ = "name", "value"
    id = "<directive>"

    def __init__(self, name, value, start_mark, end_mark):
        # type: (Any, Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.name = name
        self.value = value


class DocumentStartToken(Token):
    __slots__ = ()
    id = "<document start>"


class DocumentEndToken(Token):
    __slots__ = ()
    id = "<document end>"


class StreamStartToken(Token):
    __slots__ = ("encoding",)
    id = "<stream start>"

    def __init__(self, start_mark=None, end_mark=None, encoding=None):
        # type: (Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.encoding = encoding


class StreamEndToken(Token):
    __slots__ = ()
    id = "<stream end>"


class BlockSequenceStartToken(Token):
    __slots__ = ()
    id = "<block sequence start>"


class BlockMappingStartToken(Token):
    __slots__ = ()
    id = "<block mapping start>"


class BlockEndToken(Token):
    __slots__ = ()
    id = "<block end>"


class FlowSequenceStartToken(Token):
    __slots__ = ()
    id = "["


class FlowMappingStartToken(Token):
    __slots__ = ()
    id = "{"


class FlowSequenceEndToken(Token):
    __slots__ = ()
    id = "]"


class FlowMappingEndToken(Token):
    __slots__ = ()
    id = "}"


class KeyToken(Token):
    __slots__ = ()
    id = "?"

    # def x__repr__(self):
    #     return 'KeyToken({})'.format(
    #         self.start_mark.buffer[self.start_mark.index:].split(None, 1)[0])


class ValueToken(Token):
    __slots__ = ()
    id = ":"


class BlockEntryToken(Token):
    __slots__ = ()
    id = "-"


class FlowEntryToken(Token):
    __slots__ = ()
    id = ","


class AliasToken(Token):
    __slots__ = ("value",)
    id = "<alias>"

    def __init__(self, value, start_mark, end_mark):
        # type: (Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.value = value


class AnchorToken(Token):
    __slots__ = ("value",)
    id = "<anchor>"

    def __init__(self, value, start_mark, end_mark):
        # type: (Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.value = value


class TagToken(Token):
    __slots__ = ("value",)
    id = "<tag>"

    def __init__(self, value, start_mark, end_mark):
        # type: (Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.value = value


class ScalarToken(Token):
    __slots__ = "value", "plain", "style"
    id = "<scalar>"

    def __init__(self, value, plain, start_mark, end_mark, style=None):
        # type: (Any, Any, Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.value = value
        self.plain = plain
        self.style = style


class CommentToken(Token):
    __slots__ = "value", "pre_done"
    id = "<comment>"

    def __init__(self, value, start_mark, end_mark):
        # type: (Any, Any, Any) -> None
        Token.__init__(self, start_mark, end_mark)
        self.value = value

    def reset(self):
        # type: () -> None
        if hasattr(self, "pre_done"):
            delattr(self, "pre_done")

    def __repr__(self):
        # type: () -> Any
        v = "{!r}".format(self.value)
        if SHOWLINES:
            try:
                v += ", line: " + str(self.start_mark.line)
                v += ", col: " + str(self.start_mark.column)
            except:  # NOQA
                pass
        return "CommentToken({})".format(v)

    def __eq__(self, other):
        # type: (Any) -> bool
        if self.start_mark != other.start_mark:
            return False
        if self.end_mark != other.end_mark:
            return False
        if self.value != other.value:
            return False
        return True

    def __ne__(self, other):
        # type: (Any) -> bool
        return not self.__eq__(other)



from strictyaml.ruamel.comments import CommentedSeq, CommentedMap
from strictyaml.compound import FixedSeq, Map
from strictyaml.validators import Validator
from strictyaml.exceptions import YAMLSerializationError
from strictyaml.scalar import Bool, EmptyDict, EmptyList, Float, Int, Str


def schema_from_document(document):
    if isinstance(document, CommentedMap):
        return Map(
            {key: schema_from_document(value) for key, value in document.items()}
        )
    elif isinstance(document, CommentedSeq):
        return FixedSeq([schema_from_document(item) for item in document])
    else:
        return Str()


def schema_from_data(data, allow_empty):
    if isinstance(data, dict):
        if len(data) == 0:
            if allow_empty:
                return EmptyDict()
            raise YAMLSerializationError(
                "Empty dicts are not serializable to StrictYAML unless schema is used."
            )
        return Map(
            {key: schema_from_data(value, allow_empty) for key, value in data.items()}
        )
    elif isinstance(data, list):
        if len(data) == 0:
            if allow_empty:
                return EmptyList()
            raise YAMLSerializationError(
                "Empty lists are not serializable to StrictYAML unless schema is used."
            )
        return FixedSeq([schema_from_data(item, allow_empty) for item in data])
    elif isinstance(data, bool):
        return Bool()
    elif isinstance(data, int):
        return Int()
    elif isinstance(data, float):
        return Float()
    else:
        return Str()


class Any(Validator):
    """
    Validates any YAML and returns simple dicts/lists of strings.
    """

    def validate(self, chunk):
        return schema_from_document(chunk.contents)(chunk)

    def to_yaml(self, data, allow_empty=False):
        """
        Args:
            allow_empty (bool): True to allow EmptyDict and EmptyList in the
                    schema generated from the data.
        """
        return schema_from_data(data, allow_empty=allow_empty).to_yaml(data)

    @property
    def key_validator(self):
        return Str()



from strictyaml.exceptions import YAMLValidationError, YAMLSerializationError
from strictyaml.exceptions import InvalidValidatorError
from strictyaml.representation import YAML
from strictyaml import utils
import sys


if sys.version_info[0] == 3:
    unicode = str


class Validator(object):
    def __or__(self, other):
        return OrValidator(self, other)

    def __call__(self, chunk):
        self.validate(chunk)
        return YAML(chunk, validator=self)

    def __repr__(self):
        return "{0}()".format(self.__class__.__name__)


class MapValidator(Validator):
    def _should_be_mapping(self, data):
        if not isinstance(data, dict):
            raise YAMLSerializationError("Expected a dict, found '{}'".format(data))
        if len(data) == 0:
            raise YAMLSerializationError(
                (
                    "Expected a non-empty dict, found an empty dict.\n"
                    "Use EmptyDict validator to serialize empty dicts."
                )
            )


class SeqValidator(Validator):
    def _should_be_list(self, data):
        if not isinstance(data, list):
            raise YAMLSerializationError("Expected a list, found '{}'".format(data))
        if len(data) == 0:
            raise YAMLSerializationError(
                (
                    "Expected a non-empty list, found an empty list.\n"
                    "Use EmptyList validator to serialize empty lists."
                )
            )


class OrValidator(Validator):
    def __init__(self, validator_a, validator_b):
        assert isinstance(validator_a, Validator), "validator_a must be a Validator"
        assert isinstance(validator_b, Validator), "validator_b must be a Validator"

        self._validator_a = validator_a
        self._validator_b = validator_b

        def unpacked(validator):
            if isinstance(validator, OrValidator):
                return [
                    unpacked(validator._validator_a),
                    unpacked(validator._validator_b),
                ]
            else:
                return [validator]

        map_validator_count = len(
            [
                validator
                for validator in list(utils.flatten(unpacked(self)))
                if isinstance(validator, MapValidator)
            ]
        )

        if map_validator_count > 1:
            raise InvalidValidatorError(
                (
                    "You tried to Or ('|') together {} Map validators. "
                    "Try using revalidation instead."
                ).format(map_validator_count)
            )

        seq_validator_count = len(
            [
                validator
                for validator in list(utils.flatten(unpacked(self)))
                if isinstance(validator, SeqValidator)
            ]
        )

        if seq_validator_count > 1:
            raise InvalidValidatorError(
                (
                    "You tried to Or ('|') together {} Seq validators. "
                    "Try using revalidation instead."
                ).format(seq_validator_count)
            )

    def to_yaml(self, value):
        try:
            return self._validator_a.to_yaml(value)
        except YAMLSerializationError:
            return self._validator_b.to_yaml(value)

    def __call__(self, chunk):
        try:
            result = self._validator_a(chunk)
            result._selected_validator = result._validator
            result._validator = self
            return result
        except YAMLValidationError:
            result = self._validator_b(chunk)
            result._selected_validator = result._validator
            result._validator = self
            return result

    def __repr__(self):
        return "{0} | {1}".format(repr(self._validator_a), repr(self._validator_b))



if False:  # MYPY
    from typing import Any, Dict, Optional, List, Union, Optional, Iterator  # NOQA

anchor_attrib = "_yaml_anchor"


class Anchor(object):
    __slots__ = "value", "always_dump"
    attrib = anchor_attrib

    def __init__(self):
        # type: () -> None
        self.value = None
        self.always_dump = False

    def __repr__(self):
        # type: () -> Any
        ad = ", (always dump)" if self.always_dump else ""
        return "Anchor({!r}{})".format(self.value, ad)



from strictyaml.ruamel.comments import CommentedSeq, CommentedMap
from strictyaml.ruamel import dump, RoundTripDumper
from copy import deepcopy
import sys


if sys.version_info[0] == 3:
    unicode = str


class YAMLPointer(object):
    """
    A sequence of indexes/keys that look up a specific chunk of a YAML document.

    A YAML pointer can point to a key, value, item in a sequence or part of a string
    in a value or item.
    """

    def __init__(self):
        self._indices = []

    @property
    def last_index(self):
        assert self.is_index()
        return self._indices[-1][1]

    @property
    def last_val(self):
        assert self.is_val()
        return self._indices[-1][1]

    @property
    def last_strictkey(self):
        assert self.is_key() or self.is_val()
        return self._indices[-1][1][1]

    @property
    def last_regularkey(self):
        assert self.is_key() or self.is_val()
        return self._indices[-1][1][0]

    def val(self, regularkey, strictkey):
        assert isinstance(regularkey, (str, unicode)), type(regularkey)
        assert isinstance(strictkey, (str, unicode)), type(strictkey)
        new_location = deepcopy(self)
        new_location._indices.append(("val", (regularkey, strictkey)))
        return new_location

    def is_val(self):
        return self._indices[-1][0] == "val"

    def key(self, regularkey, strictkey):
        assert isinstance(regularkey, (str, unicode)), type(regularkey)
        assert isinstance(strictkey, (str, unicode)), type(strictkey)
        new_location = deepcopy(self)
        new_location._indices.append(("key", (regularkey, strictkey)))
        return new_location

    def is_key(self):
        return self._indices[-1][0] == "key"

    def index(self, index):
        new_location = deepcopy(self)
        new_location._indices.append(("index", index))
        return new_location

    def is_index(self):
        return self._indices[-1][0] == "index"

    def textslice(self, start, end):
        new_location = deepcopy(self)
        new_location._indices.append(("textslice", (start, end)))
        return new_location

    def is_textslice(self):
        return self._indices[-1][0] == "textslice"

    def parent(self):
        new_location = deepcopy(self)
        new_location._indices = new_location._indices[:-1]
        return new_location

    def make_child_of(self, pointer):
        new_indices = deepcopy(pointer._indices)
        new_indices.extend(self._indices)

    def _slice_segment(self, indices, segment, include_selected):
        slicedpart = deepcopy(segment)

        if len(indices) == 0 and not include_selected:
            slicedpart = None
        else:
            if len(indices) > 0:
                if indices[0][0] in ("val", "key"):
                    index = indices[0][1][0]
                else:
                    index = indices[0][1]
                start_popping = False

                if isinstance(segment, CommentedMap):
                    for key in segment.keys():
                        if start_popping:
                            slicedpart.pop(key)

                        if index == key:
                            start_popping = True

                            if isinstance(segment[index], (CommentedSeq, CommentedMap)):
                                slicedpart[index] = self._slice_segment(
                                    indices[1:],
                                    segment[index],
                                    include_selected=include_selected,
                                )

                            if not include_selected and len(indices) == 1:
                                slicedpart.pop(key)

                if isinstance(segment, CommentedSeq):
                    for i, value in enumerate(segment):
                        if start_popping:
                            del slicedpart[-1]

                        if i == index:
                            start_popping = True

                            if isinstance(segment[index], (CommentedSeq, CommentedMap)):
                                slicedpart[index] = self._slice_segment(
                                    indices[1:],
                                    segment[index],
                                    include_selected=include_selected,
                                )

                            if not include_selected and len(indices) == 1:
                                slicedpart.pop(index)

        return slicedpart

    def start_line(self, document):
        slicedpart = self._slice_segment(
            self._indices, document, include_selected=False
        )

        if slicedpart is None or slicedpart == {} or slicedpart == []:
            return 1
        else:
            return (
                len(dump(slicedpart, Dumper=RoundTripDumper).rstrip().split("\n")) + 1
            )

    def end_line(self, document):
        slicedpart = self._slice_segment(self._indices, document, include_selected=True)
        return len(dump(slicedpart, Dumper=RoundTripDumper).rstrip().split("\n"))

    def lines(self, document):
        return "\n".join(
            dump(document, Dumper=RoundTripDumper).split("\n")[
                self.start_line(document) - 1 : self.end_line(document)
            ]
        )

    def lines_before(self, document, how_many):
        return "\n".join(
            dump(document, Dumper=RoundTripDumper).split("\n")[
                self.start_line(document) - 1 - how_many : self.start_line(document) - 1
            ]
        )

    def lines_after(self, document, how_many):
        return "\n".join(
            dump(document, Dumper=RoundTripDumper).split("\n")[
                self.end_line(document) : self.end_line(document) + how_many
            ]
        )

    def _individual_get(self, segment, index_type, index, strictdoc):
        if index_type == "val":
            for key, value in segment.items():
                if key == index[0]:
                    return value
                if hasattr(key, "text"):
                    if key.text == index[0]:
                        return value
            raise Exception("Invalid state")
        elif index_type == "index":
            return segment[index]
        elif index_type == "textslice":
            return segment[index[0] : index[1]]
        elif index_type == "key":
            return index[1] if strictdoc else index[0]
        else:
            raise Exception("Invalid state")

    def get(self, document, strictdoc=False):
        segment = document
        for index_type, index in self._indices:
            segment = self._individual_get(segment, index_type, index, strictdoc)
        return segment

    def set(self, src_obj, src_attr, new_ruamel, strictdoc=False):
        """Since set() needs to overwrite what this pointer points to, it
        affects the parent object.  Therefore, rather than taking "document"
        as get(), it takes the object which holds the document and the name
        of the property which is the document.
        """
        obj_last = src_obj
        key_last = src_attr
        r = getattr(src_obj, src_attr)
        for index_type, index in self._indices:
            obj_last = r
            if index_type == "val":
                key_last = index[1] if strictdoc else index[0]
                r = r[key_last]
            elif index_type == "index":
                key_last = index
                r = r[key_last]
            elif index_type == "textslice":
                key_last = None
                r = r[index[0] : index[1]]
            elif index_type == "key":
                key_last = None
                r = index[1] if strictdoc else index[0]
            else:
                raise RuntimeError("Invalid state")
        if obj_last is src_obj:
            # Starts with an attribute set
            setattr(src_obj, src_attr, new_ruamel)
        elif key_last is not None:
            # Others are item set
            if hasattr(obj_last, "_value"):
                # Only want to overwrite value, do NOT re-validate schema...
                obj_last._value[key_last] = new_ruamel
            else:
                obj_last[key_last] = new_ruamel
        else:
            raise NotImplementedError("invalid key, cannot set")

    def __repr__(self):
        return "<YAMLPointer: {0}>".format(self._indices)
